use std::borrow::Cow;
use std::collections::{BTreeMap, HashMap, HashSet};
use std::convert::{TryFrom, TryInto};
use std::fmt;
use std::ops::Bound;
use std::sync::{atomic, Arc, RwLock};

use dataflow_expression::{BinaryOperator as DfBinaryOperator, Expr as DfExpr};
use itertools::Itertools;
use launchpad::redacted::Sensitive;
use nom_sql::analysis::visit_mut::VisitorMut;
use nom_sql::{
    self, BinaryOperator, ColumnConstraint, DeleteStatement, Expr, InsertStatement, Literal,
    Relation, SelectStatement, SqlIdentifier, SqlQuery, UnaryOperator, UpdateStatement,
};
use readyset::consistency::Timestamp;
use readyset::internal::LocalNodeIndex;
use readyset::recipe::changelist::{Change, ChangeList, IntoChanges};
use readyset::results::{ResultIterator, Results};
use readyset::{
    ColumnSchema, KeyColumnIdx, KeyComparison, ReadQuery, ReaderAddress, ReadySetError,
    ReadySetHandle, ReadySetResult, SchemaType, Table, TableOperation, View, ViewCreateRequest,
    ViewPlaceholder, ViewQuery, ViewSchema,
};
use readyset_data::{DfType, DfValue, Dialect};
use readyset_errors::ReadySetError::PreparedStatementMissing;
use readyset_errors::{
    internal, internal_err, invariant_eq, table_err, unsupported, unsupported_err,
};
use readyset_server::worker::readers::{CallResult, ReadRequestHandler};
use readyset_sql_passes::anonymize::anonymize_literals;
use tracing::{error, info, instrument, trace};
use vec1::vec1;

use crate::backend::SelectSchema;
use crate::rewrite::{self, ProcessedQueryParams};
use crate::utils;

type StatementID = u32;

#[derive(Clone)]
pub(crate) enum PreparedStatement {
    Select(PreparedSelectStatement),
    Insert(nom_sql::InsertStatement),
    Update(nom_sql::UpdateStatement),
    Delete(DeleteStatement),
}

#[derive(Clone)]
pub(crate) struct PreparedSelectStatement {
    name: Relation,
    statement: Box<nom_sql::SelectStatement>,
    processed_query_params: ProcessedQueryParams,
}

impl fmt::Debug for PreparedStatement {
    fn fmt(&self, f: &mut fmt::Formatter) -> Result<(), fmt::Error> {
        match self {
            PreparedStatement::Select(PreparedSelectStatement {
                name, statement, ..
            }) => write!(f, "{}: {}", name, statement),
            PreparedStatement::Insert(s) => write!(f, "{}", s),
            PreparedStatement::Update(s) => write!(f, "{}", s),
            PreparedStatement::Delete(s) => write!(f, "{}", s),
        }
    }
}

/// Wrapper around a NoriaBackendInner which may not have been successfully
/// created. When this is the case, this wrapper allows returning an error
/// from any call that requires NoriaBackendInner through an error
/// returned by `get_mut`.
pub struct NoriaBackend {
    inner: Option<NoriaBackendInner>,
}

impl NoriaBackend {
    fn get_mut(&mut self) -> ReadySetResult<&mut NoriaBackendInner> {
        // TODO(ENG-707): Support retrying to create a backend in the future.
        self.inner
            .as_mut()
            .ok_or_else(|| internal_err!("Failed to create a Noria backend."))
    }
}

pub struct NoriaBackendInner {
    noria: ReadySetHandle,
    tables: BTreeMap<Relation, Table>,
    views: BTreeMap<Relation, View>,
    /// The server can handle (non-parameterized) LIMITs and (parameterized) OFFSETs in the
    /// dataflow graph
    server_supports_pagination: bool,
}

macro_rules! noria_await {
    ($self:expr, $fut:expr) => {{
        let noria = &mut $self.noria;

        futures_util::future::poll_fn(|cx| noria.poll_ready(cx)).await?;
        $fut.await
    }};
}

impl NoriaBackendInner {
    async fn new(ch: ReadySetHandle, server_supports_pagination: bool) -> Self {
        NoriaBackendInner {
            tables: BTreeMap::new(),
            views: BTreeMap::new(),
            noria: ch,
            server_supports_pagination,
        }
    }

    async fn get_noria_table(&mut self, table: &Relation) -> ReadySetResult<&mut Table> {
        if !self.tables.contains_key(table) {
            let t = noria_await!(self, self.noria.table(table.clone()))?;
            self.tables.insert(table.to_owned(), t);
        }
        Ok(self.tables.get_mut(table).unwrap())
    }

    /// If `invalidate_cache` is passed, the view cache, `views` will be ignored and a view will be
    /// retrieved from noria.
    async fn get_noria_view<'a>(
        &'a mut self,
        view: &Relation,
        invalidate_cache: bool,
    ) -> ReadySetResult<&'a mut View> {
        if invalidate_cache {
            self.views.remove(view);
        }
        if !self.views.contains_key(view) {
            let vh = noria_await!(self, self.noria.view(view.clone()))?;
            self.views.insert(view.to_owned(), vh);
        }
        Ok(self.views.get_mut(view).unwrap())
    }
}

#[derive(Debug, Clone)]
pub enum PrepareResult {
    Select {
        statement_id: u32,
        params: Vec<ColumnSchema>,
        schema: Vec<ColumnSchema>,
    },
    Insert {
        statement_id: u32,
        params: Vec<ColumnSchema>,
        schema: Vec<ColumnSchema>,
    },
    Update {
        statement_id: u32,
        params: Vec<ColumnSchema>,
    },
    Delete {
        statement_id: u32,
        params: Vec<ColumnSchema>,
    },
}

impl PrepareResult {
    /// Get the noria statement id for this statement
    pub fn statement_id(&self) -> u32 {
        match self {
            PrepareResult::Select { statement_id, .. }
            | PrepareResult::Insert { statement_id, .. }
            | PrepareResult::Delete { statement_id, .. }
            | PrepareResult::Update { statement_id, .. } => *statement_id,
        }
    }
}

/// A single row in the variable table associated with [`QueryResult::MetaVariables`].
#[derive(Debug)]
pub struct MetaVariable {
    /// The variable name.
    pub name: SqlIdentifier,
    /// The value associated with the variable.
    pub value: String,
}

impl<N: Into<SqlIdentifier>, V: Into<String>> From<(N, V)> for MetaVariable {
    fn from((name, value): (N, V)) -> Self {
        MetaVariable {
            name: name.into(),
            value: value.into(),
        }
    }
}

#[derive(Debug)]
#[allow(clippy::large_enum_variant)]
pub enum QueryResult<'a> {
    Empty,
    Insert {
        num_rows_inserted: u64,
        first_inserted_id: u64,
    },
    Select {
        rows: ResultIterator,
        schema: SelectSchema<'a>,
    },
    Update {
        num_rows_updated: u64,
        last_inserted_id: u64,
    },
    Delete {
        num_rows_deleted: u64,
    },
    /// A metadata table returned as a response to eg an EXPLAIN query. Unlike
    /// [`QueryResult::MetaVariables`] it will format the output as a table with a single row,
    /// where the columns names correspond to the [`MetaVariable`] names.
    Meta(Vec<MetaVariable>),
    /// A table of variables returned as a response to a SHOW READYSET STATUS query.
    MetaVariables(Vec<MetaVariable>),
    /// A table of variables returned as a response to a SHOW READYSET STATUS query.
    /// The first MetaVariable serves as the column headers
    MetaWithHeader(Vec<MetaVariable>),
}

impl<'a> QueryResult<'a> {
    pub fn from_owned(schema: SelectSchema<'a>, data: Vec<Results>) -> Self {
        QueryResult::Select {
            schema,
            rows: ResultIterator::owned(data),
        }
    }

    pub fn empty(schema: SelectSchema<'a>) -> Self {
        QueryResult::Select {
            schema,
            rows: ResultIterator::owned(vec![]),
        }
    }

    pub fn from_iter(schema: SelectSchema<'a>, rows: ResultIterator) -> Self {
        QueryResult::Select { schema, rows }
    }

    #[inline]
    pub fn into_owned(self) -> QueryResult<'static> {
        match self {
            QueryResult::Select { schema, rows } => QueryResult::Select {
                schema: schema.into_owned(),
                rows,
            },
            // Have to manually pass each variant to convince rustc that the
            // returned type is really owned
            QueryResult::Empty => QueryResult::Empty,
            QueryResult::Insert {
                num_rows_inserted,
                first_inserted_id,
            } => QueryResult::Insert {
                num_rows_inserted,
                first_inserted_id,
            },
            QueryResult::Update {
                num_rows_updated,
                last_inserted_id,
            } => QueryResult::Update {
                num_rows_updated,
                last_inserted_id,
            },
            QueryResult::Delete { num_rows_deleted } => QueryResult::Delete { num_rows_deleted },
            QueryResult::Meta(meta) => QueryResult::Meta(meta),
            QueryResult::MetaVariables(vec) => QueryResult::MetaVariables(vec),
            QueryResult::MetaWithHeader(vec) => QueryResult::MetaWithHeader(vec),
        }
    }
}

#[derive(Clone)]
pub struct ViewCache {
    /// Global cache of view endpoints and prepared statements.
    global: Arc<RwLock<HashMap<ViewCreateRequest, Relation>>>,
    /// Thread-local version of global cache (consulted first).
    local: HashMap<ViewCreateRequest, Relation>,
}

impl ViewCache {
    /// Construct a new ViewCache with a passed in global view cache.
    pub fn new(global_cache: Arc<RwLock<HashMap<ViewCreateRequest, Relation>>>) -> ViewCache {
        ViewCache {
            global: global_cache,
            local: HashMap::new(),
        }
    }

    /// Registers a statement with the provided name into both the local and global view caches.
    pub fn register_statement(&mut self, name: &Relation, view_request: ViewCreateRequest) {
        self.local
            .entry(view_request.clone())
            .or_insert_with(|| name.clone());
        tokio::task::block_in_place(move || {
            self.global
                .write()
                .unwrap()
                .entry(view_request)
                .or_insert_with(|| name.clone());
        });
    }

    /// Retrieves the name for the provided statement if it's in the cache. We first check local
    /// cache, and if it's not there we check global cache. If it's in global but not local, we
    /// backfill local cache before returning the result.
    pub fn statement_name(&mut self, view_request: &ViewCreateRequest) -> Option<Relation> {
        let maybe_name = if let Some(name) = self.local.get(view_request) {
            return Some(name.clone());
        } else {
            // Didn't find it in local, so let's check global.
            let gc = tokio::task::block_in_place(|| self.global.read().unwrap());
            gc.get(view_request).cloned()
        };

        maybe_name.map(|n| {
            // Backfill into local before we return.
            self.local.insert(view_request.clone(), n.clone());
            n
        })
    }

    /// Removes the statement with the given name from both the global and local caches.
    pub fn remove_statement(&mut self, name: &Relation) {
        self.local.retain(|_, v| v != name);
        tokio::task::block_in_place(|| {
            self.global.write().unwrap().retain(|_, v| v != name);
        });
    }

    /// Clears all statements from all caches
    fn clear(&mut self) {
        self.local.clear();
        tokio::task::block_in_place(|| {
            self.global.write().unwrap().clear();
        })
    }

    /// Returns the original view create request based on a provided name if it exists in either the
    /// local or global caches.
    pub fn view_create_request_from_name(&self, name: &Relation) -> Option<ViewCreateRequest> {
        self.local
            .iter()
            .find(|(_, n)| *n == name)
            .map(|(v, _)| v.clone())
            .or_else(|| {
                tokio::task::block_in_place(|| {
                    self.global
                        .read()
                        .unwrap()
                        .iter()
                        .find(|(_, n)| *n == name)
                        .map(|(v, _)| v.clone())
                })
            })
    }
}

pub struct NoriaConnector {
    inner: NoriaBackend,
    auto_increments: Arc<RwLock<HashMap<Relation, atomic::AtomicUsize>>>,
    /// Global and thread-local cache of view endpoints and prepared statements.
    view_cache: ViewCache,

    prepared_statement_cache: HashMap<StatementID, PreparedStatement>,

    /// Set of views that have failed on previous requests. Separate from the backend
    /// to allow returning references to schemas from views all the way to mysql-srv,
    /// but on subsequent requests, do not use a failed view.
    failed_views: HashSet<Relation>,

    /// How to handle issuing reads against ReadySet. See [`ReadBehavior`].
    read_behavior: ReadBehavior,

    /// A read request handler that may be used to service reads from readers
    /// on the same server.
    read_request_handler: request_handler::LocalReadHandler,

    /// SQL Dialect to pass to ReadySet as part of all migration requests
    dialect: Dialect,

    /// Currently configured search path for schemas.
    ///
    /// Note that the terminology used here is maximally general - while only PostgreSQL truly
    /// supports a multi-element schema search path, the concept of "currently connected database"
    /// in MySQL can be thought of as a schema search path that only has one element.
    schema_search_path: Vec<SqlIdentifier>,
}

mod request_handler {
    use readyset_server::worker::readers::ReadRequestHandler;

    /// Since [`ReadRequestHandler`] contains some fields that aren't [`Sync`], this is a workaround
    /// wrapper to make it safely [`Sync`], by ensuring that all accesses to the underlying
    /// [`ReadRequestHandler`] are exclusive. This effectively makes sure that no references are
    /// shared between threads ever. This is implemented in a submodule so that the private fields
    /// are not accidentally accessed without an exclusive reference.
    #[derive(Clone)]
    #[repr(transparent)]
    pub(super) struct LocalReadHandler(Option<ReadRequestHandler>);

    impl LocalReadHandler {
        pub(super) fn new(handler: Option<ReadRequestHandler>) -> Self {
            LocalReadHandler(handler)
        }

        #[inline]
        pub(super) fn as_mut(&mut self) -> Option<&mut ReadRequestHandler> {
            self.0.as_mut()
        }
    }

    /// SAFETY: since all accesses to the inner field are exclusive, no references are ever shared
    /// between threads.
    unsafe impl Sync for LocalReadHandler {}
}

/// The read behavior used when executing a read against ReadySet.
#[derive(Clone, Copy)]
pub enum ReadBehavior {
    /// If ReadySet is unable to immediately service the read due to a cache miss, block on the
    /// response.
    Blocking,
    /// If ReadySet is unable to immediately service the read, return an error.
    NonBlocking,
}

impl ReadBehavior {
    fn is_blocking(&self) -> bool {
        matches!(self, Self::Blocking)
    }
}

/// Used when we can determine that the params for 'OFFSET ?' or 'LIMIT ?' passed in
/// with an execute statement will result in an empty resultset
async fn short_circuit_empty_resultset(getter: &mut View) -> ReadySetResult<QueryResult<'_>> {
    let getter_schema = getter
        .schema()
        .ok_or_else(|| internal_err!("No schema for view"))?;
    Ok(QueryResult::empty(SelectSchema {
        use_bogo: false,
        schema: Cow::Borrowed(getter_schema.schema(SchemaType::ReturnedSchema)),
        columns: Cow::Borrowed(getter.columns()),
    }))
}

/// Provides the necessary context to execute a select statement against noria, either for a
/// prepared or an ad-hoc query
#[allow(clippy::large_enum_variant)]
#[derive(Debug)]
pub(crate) enum ExecuteSelectContext<'ctx> {
    Prepared {
        q_id: u32,
        params: &'ctx [DfValue],
    },
    AdHoc {
        statement: nom_sql::SelectStatement,
        query: &'ctx str,
        create_if_missing: bool,
    },
}

impl NoriaConnector {
    pub async fn new(
        ch: ReadySetHandle,
        auto_increments: Arc<RwLock<HashMap<Relation, atomic::AtomicUsize>>>,
        query_cache: Arc<RwLock<HashMap<ViewCreateRequest, Relation>>>,
        read_behavior: ReadBehavior,
        dialect: Dialect,
        schema_search_path: Vec<SqlIdentifier>,
        server_supports_pagination: bool,
    ) -> Self {
        NoriaConnector::new_with_local_reads(
            ch,
            auto_increments,
            query_cache,
            read_behavior,
            None,
            dialect,
            schema_search_path,
            server_supports_pagination,
        )
        .await
    }

    #[allow(clippy::too_many_arguments)]
    pub async fn new_with_local_reads(
        ch: ReadySetHandle,
        auto_increments: Arc<RwLock<HashMap<Relation, atomic::AtomicUsize>>>,
        query_cache: Arc<RwLock<HashMap<ViewCreateRequest, Relation>>>,
        read_behavior: ReadBehavior,
        read_request_handler: Option<ReadRequestHandler>,
        dialect: Dialect,
        schema_search_path: Vec<SqlIdentifier>,
        server_supports_pagination: bool,
    ) -> Self {
        let backend = NoriaBackendInner::new(ch, server_supports_pagination).await;

        NoriaConnector {
            inner: NoriaBackend {
                inner: Some(backend),
            },
            auto_increments,
            view_cache: ViewCache::new(query_cache),
            prepared_statement_cache: HashMap::new(),
            failed_views: HashSet::new(),
            read_behavior,
            read_request_handler: request_handler::LocalReadHandler::new(read_request_handler),
            dialect,
            schema_search_path,
        }
    }

    pub(crate) async fn graphviz(
        &mut self,
        simplified: bool,
    ) -> ReadySetResult<QueryResult<'static>> {
        let noria = &mut self.inner.get_mut()?.noria;

        let (label, graphviz) = if simplified {
            ("SIMPLIFIED GRAPHVIZ", noria.simple_graphviz().await?)
        } else {
            ("GRAPHVIZ", noria.graphviz().await?)
        };

        Ok(QueryResult::Meta(vec![(label, graphviz).into()]))
    }

    pub(crate) async fn verbose_views(&mut self) -> ReadySetResult<QueryResult<'static>> {
        let noria = &mut self.inner.get_mut()?.noria;
        let views = noria.verbose_views().await?;
        //TODO(DAN): this is ridiculous, update Meta instead
        let select_schema = SelectSchema {
            use_bogo: false,
            schema: Cow::Owned(vec![
                ColumnSchema {
                    column: nom_sql::Column {
                        name: "name".into(),
                        table: None,
                    },
                    column_type: DfType::DEFAULT_TEXT,
                    base: None,
                },
                ColumnSchema {
                    column: nom_sql::Column {
                        name: "query".into(),
                        table: None,
                    },
                    column_type: DfType::DEFAULT_TEXT,
                    base: None,
                },
                ColumnSchema {
                    column: nom_sql::Column {
                        name: "fallback behavior".into(),
                        table: None,
                    },
                    column_type: DfType::DEFAULT_TEXT,
                    base: None,
                },
            ]),

            columns: Cow::Owned(vec![
                "name".into(),
                "query".into(),
                "fallback behavior".into(),
            ]),
        };
        let data = views
            .into_iter()
            .map(|(n, (mut q, always))| {
                anonymize_literals(&mut q);
                vec![
                    DfValue::from(n.to_string()),
                    DfValue::from(q.to_string()),
                    DfValue::from(if always {
                        "no fallback"
                    } else {
                        "fallback allowed"
                    }),
                ]
            })
            .collect::<Vec<_>>();
        Ok(QueryResult::from_owned(
            select_schema,
            vec![Results::new(data)],
        ))
    }

    pub(crate) fn server_supports_pagination(&self) -> bool {
        self.inner
            .inner
            .as_ref()
            .map(|v| v.server_supports_pagination)
            .unwrap_or(false)
    }

    // TODO(andrew): Allow client to map table names to NodeIndexes without having to query ReadySet
    // repeatedly. Eventually, this will be responsibility of the TimestampService.
    pub async fn node_index_of(&mut self, table_name: &str) -> ReadySetResult<LocalNodeIndex> {
        let table_handle = self.inner.get_mut()?.noria.table(table_name).await?;
        Ok(table_handle.node)
    }

    pub async fn handle_insert(
        &mut self,
        q: &nom_sql::InsertStatement,
    ) -> ReadySetResult<QueryResult<'_>> {
        let table = &q.table;

        // create a mutator if we don't have one for this table already
        trace!(%table, "query::insert::access mutator");
        let putter = self.inner.get_mut()?.get_noria_table(table).await?;
        trace!("query::insert::extract schema");
        let schema = putter
            .schema()
            .ok_or_else(|| internal_err!("no schema for table '{}'", table))?;

        // set column names (insert schema) if not set
        let q = match q.fields {
            Some(_) => Cow::Borrowed(q),
            None => {
                let mut query = q.clone();
                query.fields = Some(schema.fields.iter().map(|cs| cs.column.clone()).collect());
                Cow::Owned(query)
            }
        };

        let data: Vec<Vec<DfValue>> = q
            .data
            .iter()
            .map(|row| {
                row.iter()
                    .map(|expr| match expr {
                        Expr::Literal(lit) => DfValue::try_from(lit),
                        // Ad-hoc handle unary negation (for logictests, to allow them to insert
                        // negative values)
                        Expr::UnaryOp {
                            op: UnaryOperator::Neg,
                            rhs: box Expr::Literal(lit),
                        } => {
                            let val = DfValue::try_from(lit)?;
                            &val * &(-1).into()
                        }
                        _ => unsupported!("Only literal values are supported in expressions"),
                    })
                    .collect::<Result<Vec<_>, _>>()
            })
            .collect::<Result<Vec<_>, _>>()?;

        self.do_insert(&q, data).await
    }

    pub async fn prepare_insert(
        &mut self,
        mut q: nom_sql::InsertStatement,
        statement_id: u32,
    ) -> ReadySetResult<PrepareResult> {
        trace!(table = %q.table.name, "insert::access mutator");
        let mutator = self.inner.get_mut()?.get_noria_table(&q.table).await?;
        trace!("insert::extract schema");
        let schema = mutator
            .schema()
            .ok_or_else(|| internal_err!("Could not find schema for table {}", q.table.name))?
            .fields
            .iter()
            .map(|cs| ColumnSchema::from_base(cs.clone(), q.table.clone(), self.dialect))
            .collect::<Result<Vec<_>, _>>()?;

        if q.fields.is_none() {
            q.fields = Some(
                mutator
                    .schema()
                    .as_ref()
                    .unwrap()
                    .fields
                    .iter()
                    .map(|cs| cs.column.clone())
                    .collect(),
            );
        }

        let params: Vec<_> = {
            // extract parameter columns -- easy here, since they must all be in the same table
            let param_cols = utils::insert_statement_parameter_columns(&q);
            param_cols
                .into_iter()
                .map(|c| {
                    schema
                        .iter()
                        .cloned()
                        .find(|mc| c.name == mc.column.name)
                        .ok_or_else(|| internal_err!("column '{}' missing in mutator schema", c))
                })
                .collect::<ReadySetResult<Vec<_>>>()?
        };

        // nothing more to do for an insert
        // register a new prepared statement
        trace!(id = statement_id, "insert::registered");
        self.prepared_statement_cache
            .insert(statement_id, PreparedStatement::Insert(q));
        Ok(PrepareResult::Insert {
            statement_id,
            params,
            schema,
        })
    }

    pub(crate) async fn execute_prepared_insert(
        &mut self,
        q_id: u32,
        params: &[DfValue],
    ) -> ReadySetResult<QueryResult<'_>> {
        let prep: PreparedStatement = self
            .prepared_statement_cache
            .get(&q_id)
            .ok_or(PreparedStatementMissing { statement_id: q_id })?
            .clone();
        trace!("delegate");
        match prep {
            PreparedStatement::Insert(ref q) => {
                let table = &q.table;
                let putter = self.inner.get_mut()?.get_noria_table(table).await?;
                trace!("insert::extract schema");
                let schema = putter
                    .schema()
                    .ok_or_else(|| internal_err!("no schema for table '{}'", table))?;
                // unwrap: safe because we always pass in Some(params) so don't hit None path of
                // coerce_params
                let coerced_params = utils::coerce_params(
                    Some(params),
                    &SqlQuery::Insert(q.clone()),
                    schema,
                    self.dialect,
                )
                .unwrap()
                .unwrap();
                self.do_insert(q, vec![coerced_params]).await
            }
            _ => {
                internal!(
                    "Execute_prepared_insert is being called for a non insert prepared statement."
                )
            }
        }
    }

    pub(crate) async fn handle_delete(
        &mut self,
        q: &nom_sql::DeleteStatement,
    ) -> ReadySetResult<QueryResult<'_>> {
        let cond = q
            .where_clause
            .as_ref()
            .ok_or_else(|| unsupported_err!("only supports DELETEs with WHERE-clauses"))?;

        // create a mutator if we don't have one for this table already
        trace!(table = %q.table.name, "delete::access mutator");
        let mutator = self.inner.get_mut()?.get_noria_table(&q.table).await?;

        trace!("delete::extract schema");
        let pkey = if let Some(cts) = mutator.schema() {
            utils::get_primary_key(cts)
                .into_iter()
                .map(|(_, c)| c)
                .collect::<Vec<_>>()
        } else {
            unsupported!("cannot delete from view");
        };

        trace!("delete::flatten conditionals");
        match utils::flatten_conditional(cond, &pkey)? {
            None => Ok(QueryResult::Delete {
                num_rows_deleted: 0_u64,
            }),
            Some(ref flattened) if flattened.is_empty() => {
                unsupported!("DELETE only supports WHERE-clauses on primary keys")
            }
            Some(flattened) => {
                let count = flattened.len() as u64;
                trace!("delete::execute");
                for key in flattened {
                    if let Err(e) = mutator.delete(key).await {
                        error!(error = %e, "failed");
                        return Err(e);
                    };
                }
                trace!("delete::done");
                Ok(QueryResult::Delete {
                    num_rows_deleted: count,
                })
            }
        }
    }

    pub(crate) async fn handle_update<'a>(
        &'a mut self,
        q: &nom_sql::UpdateStatement,
    ) -> ReadySetResult<QueryResult<'a>> {
        self.do_update(Cow::Borrowed(q), None).await
    }

    pub(crate) async fn prepare_update(
        &mut self,
        q: nom_sql::UpdateStatement,
        statement_id: u32,
    ) -> ReadySetResult<PrepareResult> {
        // ensure that we have schemas and endpoints for the query
        trace!(table = %q.table.name, "update::access mutator");
        let mutator = self.inner.get_mut()?.get_noria_table(&q.table).await?;
        trace!("update::extract schema");
        let table_schema = mutator
            .schema()
            .ok_or_else(|| internal_err!("Could not find schema for table {}", q.table.name))?;

        // extract parameter columns
        let params = utils::update_statement_parameter_columns(&q)
            .into_iter()
            .map(|c| {
                table_schema
                    .fields
                    .iter()
                    // We know that only one table is mentioned, so no need to match on both table
                    // and name - just check name here
                    .find(|f| f.column.name == c.name)
                    .cloned()
                    .map(|cs| ColumnSchema::from_base(cs, q.table.clone(), self.dialect))
                    .transpose()?
                    .ok_or_else(|| internal_err!("Unknown column {}", c))
            })
            .collect::<Result<Vec<_>, _>>()?;

        trace!(id = statement_id, "update::registered");
        self.prepared_statement_cache
            .insert(statement_id, PreparedStatement::Update(q));
        Ok(PrepareResult::Update {
            statement_id,
            params,
        })
    }

    pub(crate) async fn execute_prepared_update(
        &mut self,
        q_id: u32,
        params: &[DfValue],
    ) -> ReadySetResult<QueryResult<'_>> {
        let prep: PreparedStatement = self
            .prepared_statement_cache
            .get(&q_id)
            .ok_or(PreparedStatementMissing { statement_id: q_id })?
            .clone();

        trace!("delegate");
        match prep {
            PreparedStatement::Update(q) => self.do_update(Cow::Owned(q), Some(params)).await,
            _ => internal!(),
        }
    }

    pub(crate) async fn prepare_delete(
        &mut self,
        q: DeleteStatement,
        statement_id: u32,
    ) -> ReadySetResult<PrepareResult> {
        // ensure that we have schemas and endpoints for the query
        trace!(table = %q.table.name, "delete::access mutator");
        let mutator = self.inner.get_mut()?.get_noria_table(&q.table).await?;
        trace!("delete::extract schema");
        let table_schema = mutator
            .schema()
            .ok_or_else(|| internal_err!("Could not find schema for table {}", q.table.name))?;

        // extract parameter columns
        let params = utils::delete_statement_parameter_columns(&q)
            .into_iter()
            .map(|c| {
                table_schema
                    .fields
                    .iter()
                    // We know that only one table is mentioned, so no need to match on both table
                    // and name - just check name here
                    .find(|f| f.column.name == c.name)
                    .cloned()
                    .map(|cs| ColumnSchema::from_base(cs, q.table.clone(), self.dialect))
                    .transpose()?
                    .ok_or_else(|| internal_err!("Unknown column {}", c))
            })
            .collect::<Result<Vec<_>, _>>()?;

        trace!(id = statement_id, "delete::registered");
        self.prepared_statement_cache
            .insert(statement_id, PreparedStatement::Delete(q));
        Ok(PrepareResult::Delete {
            statement_id,
            params,
        })
    }

    pub(crate) async fn execute_prepared_delete(
        &mut self,
        q_id: u32,
        params: &[DfValue],
    ) -> ReadySetResult<QueryResult<'_>> {
        let prep: PreparedStatement = self
            .prepared_statement_cache
            .get(&q_id)
            .ok_or(PreparedStatementMissing { statement_id: q_id })?
            .clone();

        trace!("delegate");
        match prep {
            PreparedStatement::Delete(q) => self.do_delete(Cow::Owned(q), Some(params)).await,
            _ => internal!(),
        }
    }

    /// Calls the `extend_recipe` endpoint on ReadySet with the given query.
    pub(crate) async fn handle_table_operation<C>(
        &mut self,
        changes: C,
    ) -> ReadySetResult<QueryResult<'_>>
    where
        C: IntoChanges,
    {
        // TODO(malte): we should perhaps check our usual caches here, rather than just blindly
        // doing a migration on ReadySet ever time. On the other hand, CREATE TABLE is rare...
        noria_await!(
            self.inner.get_mut()?,
            self.inner.get_mut()?.noria.extend_recipe(
                ChangeList::from_changes(changes, self.dialect)
                    .with_schema_search_path(self.schema_search_path.clone())
            )
        )?;
        Ok(QueryResult::Empty)
    }

    pub(crate) async fn readyset_status(&mut self) -> ReadySetResult<QueryResult<'static>> {
        let status = noria_await!(self.inner.get_mut()?, self.inner.get_mut()?.noria.status())?;

        // Converts from ReadySetStatus -> Vec<(String, String)> -> QueryResult
        Ok(QueryResult::MetaVariables(
            <Vec<(String, String)>>::from(status)
                .into_iter()
                .map(MetaVariable::from)
                .collect(),
        ))
    }

    /// Set the schema search path
    pub fn set_schema_search_path(&mut self, search_path: Vec<SqlIdentifier>) {
        self.schema_search_path = search_path;
    }

    /// Returns a reference to the currently configured schema search path
    pub fn schema_search_path(&self) -> &[SqlIdentifier] {
        self.schema_search_path.as_ref()
    }
}

impl NoriaConnector {
    /// This function handles CREATE CACHE statements. When explicit-migrations is enabled,
    /// this function is the only way to create a view in noria.
    pub async fn handle_create_cached_query(
        &mut self,
        name: Option<&Relation>,
        statement: &nom_sql::SelectStatement,
        override_schema_search_path: Option<Vec<SqlIdentifier>>,
        always: bool,
    ) -> ReadySetResult<()> {
        let name = name.cloned().unwrap_or_else(|| {
            utils::generate_query_name(statement, self.schema_search_path()).into()
        });
        let schema_search_path =
            override_schema_search_path.unwrap_or_else(|| self.schema_search_path.clone());
        let changelist = ChangeList::from_change(
            Change::create_cache(name.clone(), statement.clone(), always),
            self.dialect,
        )
        .with_schema_search_path(schema_search_path.clone());

        noria_await!(
            self.inner.get_mut()?,
            self.inner.get_mut()?.noria.extend_recipe(changelist)
        )?;

        // If the query is already in there with a different name, we don't need to make a new name
        // for it, as *lookups* only need one of the names for the query, and when we drop it we'll
        // be hitting noria anyway
        self.view_cache.register_statement(
            &name,
            ViewCreateRequest::new(statement.clone(), schema_search_path),
        );

        Ok(())
    }

    async fn get_view(
        &mut self,
        q: &nom_sql::SelectStatement,
        prepared: bool,
        create_if_not_exist: bool,
    ) -> ReadySetResult<Relation> {
        let view_request = ViewCreateRequest::new(q.clone(), self.schema_search_path.clone());
        match self.view_cache.statement_name(&view_request) {
            None => {
                let qname: Relation =
                    utils::generate_query_name(q, self.schema_search_path()).into();

                // add the query to ReadySet
                if create_if_not_exist {
                    if prepared {
                        info!(query = %Sensitive(q), name = %qname, "adding parameterized query");
                    } else {
                        info!(query = %Sensitive(q), name = %qname, "adding ad-hoc query");
                    }

                    let changelist = ChangeList::from_change(
                        Change::create_cache(qname.clone(), q.clone(), false),
                        self.dialect,
                    )
                    .with_schema_search_path(self.schema_search_path.clone());

                    if let Err(e) = noria_await!(
                        self.inner.get_mut()?,
                        self.inner.get_mut()?.noria.extend_recipe(changelist)
                    ) {
                        error!(error = %e, "add query failed");
                        return Err(e);
                    }
                } else {
                    match noria_await!(
                        self.inner.get_mut()?,
                        self.inner.get_mut()?.noria.view(qname.clone())
                    ) {
                        Ok(view) => {
                            // We should not have an entry, but if we do it's safe to overwrite
                            // since we got this information from the controller.
                            self.inner.get_mut()?.views.insert(qname.clone(), view);
                        }
                        Err(e) => {
                            error!(error = %e, "getting view from noria failed");
                            return Err(e);
                        }
                    }
                }
                self.view_cache.register_statement(&qname, view_request);

                Ok(qname)
            }
            Some(name) => Ok(name),
        }
    }

    /// Make a request to ReadySet to drop the query with the given name, and remove it from all
    /// internal state.
    pub async fn drop_view(&mut self, name: &Relation) -> ReadySetResult<()> {
        noria_await!(
            self.inner.get_mut()?,
            self.inner.get_mut()?.noria.remove_query(name)
        )?;
        self.view_cache.remove_statement(name);
        Ok(())
    }

    /// Make a request to ReadySet to drop all cached queries, and empty all internal state
    pub async fn drop_all_caches(&mut self) -> ReadySetResult<()> {
        noria_await!(
            self.inner.get_mut()?,
            self.inner.get_mut()?.noria.remove_all_queries()
        )?;
        self.view_cache.clear();
        Ok(())
    }

    pub fn view_create_request_from_name(&self, name: &Relation) -> Option<ViewCreateRequest> {
        self.view_cache.view_create_request_from_name(name)
    }

    async fn do_insert(
        &mut self,
        q: &InsertStatement,
        data: Vec<Vec<DfValue>>,
    ) -> ReadySetResult<QueryResult<'_>> {
        let table = &q.table;

        // create a mutator if we don't have one for this table already
        trace!(%table, "insert::access mutator");
        let putter = self.inner.get_mut()?.get_noria_table(table).await?;
        trace!("insert::extract schema");
        let schema = putter
            .schema()
            .ok_or_else(|| internal_err!("no schema for table '{}'", table))?;

        let columns_specified: Vec<_> = q
            .fields
            .as_ref()
            .unwrap()
            .iter()
            .cloned()
            .map(|mut c| {
                c.table = Some(q.table.clone());
                c
            })
            .collect();

        // handle auto increment
        trace!("insert::auto-increment");
        let auto_increment_columns: Vec<_> = schema
            .fields
            .iter()
            .filter(|c| c.constraints.contains(&ColumnConstraint::AutoIncrement))
            .collect();
        if auto_increment_columns.len() > 1 {
            // can only have zero or one AUTO_INCREMENT columns
            return Err(table_err(
                table.to_string(),
                ReadySetError::MultipleAutoIncrement,
            ));
        }

        let ai = &mut self.auto_increments;
        tokio::task::block_in_place(|| {
            let ai_lock = ai.read().unwrap();
            if ai_lock.get(table).is_none() {
                drop(ai_lock);
                ai.write()
                    .unwrap()
                    .entry(table.clone())
                    .or_insert_with(|| atomic::AtomicUsize::new(0));
            }
        });
        let mut buf = vec![vec![DfValue::None; schema.fields.len()]; data.len()];
        let mut first_inserted_id = None;
        tokio::task::block_in_place(|| -> ReadySetResult<_> {
            let ai_lock = ai.read().unwrap();
            let last_insert_id = &ai_lock[table];

            // handle default values
            trace!("insert::default values");
            let mut default_value_columns = vec![];
            for c in &schema.fields {
                for cc in &c.constraints {
                    if let ColumnConstraint::DefaultValue(ref def) = *cc {
                        match def {
                            Expr::Literal(v) => {
                                default_value_columns.push((c.column.clone(), v.clone()))
                            }
                            _ => {
                                unsupported!("Only literal values are supported in default values")
                            }
                        }
                    }
                }
            }

            trace!("insert::construct ops");

            for (ri, row) in data.iter().enumerate() {
                if let Some(col) = auto_increment_columns.get(0) {
                    let idx = schema
                        .fields
                        .iter()
                        .position(|f| f == *col)
                        .ok_or_else(|| {
                            table_err(
                                table,
                                ReadySetError::NoSuchColumn(col.column.name.to_string()),
                            )
                        })?;
                    // query can specify an explicit AUTO_INCREMENT value
                    if !columns_specified.contains(&col.column) {
                        let id = last_insert_id.fetch_add(1, atomic::Ordering::SeqCst) as i64 + 1;
                        if first_inserted_id.is_none() {
                            first_inserted_id = Some(id);
                        }
                        buf[ri][idx] = DfValue::from(id);
                    }
                }

                for (c, v) in default_value_columns.drain(..) {
                    let idx = schema
                        .fields
                        .iter()
                        .position(|f| f.column == c)
                        .ok_or_else(|| {
                            table_err(table, ReadySetError::NoSuchColumn(c.name.to_string()))
                        })?;
                    // only use default value if query doesn't specify one
                    if !columns_specified.contains(&c) {
                        buf[ri][idx] = v.try_into()?;
                    }
                }

                for (ci, c) in columns_specified.iter().enumerate() {
                    let (idx, field) = schema
                        .fields
                        .iter()
                        .find_position(|f| f.column == *c)
                        .ok_or_else(|| {
                            table_err(
                                &schema.table.name,
                                ReadySetError::NoSuchColumn(c.name.to_string()),
                            )
                        })?;

                    let target_type =
                        DfType::from_sql_type(&field.sql_type, self.dialect, |_| None)?;

                    let value = row
                        .get(ci)
                        .ok_or_else(|| {
                            internal_err!(
                                "Row returned from readyset-server had the wrong number of columns",
                            )
                        })?
                        .coerce_to(&target_type, &DfType::Unknown)?; // No from_ty, we're inserting literals
                    buf[ri][idx] = value;
                }
            }
            Ok(())
        })?;

        let result = if let Some(ref update_fields) = q.on_duplicate {
            trace!("insert::complex");
            invariant_eq!(buf.len(), 1);

            let updates = {
                // fake out an update query
                let mut uq = UpdateStatement {
                    table: table.clone(),
                    fields: update_fields.clone(),
                    where_clause: None,
                };
                utils::extract_update_params_and_fields(
                    &mut uq,
                    &mut None::<std::iter::Empty<DfValue>>,
                    schema,
                    self.dialect,
                )?
            };

            // TODO(malte): why can't I consume buf here?
            let r = putter.insert_or_update(buf[0].clone(), updates).await;
            trace!("insert::complex::complete");
            r
        } else {
            trace!("insert::simple");
            let buf: Vec<_> = buf.into_iter().map(TableOperation::Insert).collect();
            let r = putter.perform_all(buf).await;
            trace!("insert::simple::complete");
            r
        };
        result?;
        Ok(QueryResult::Insert {
            num_rows_inserted: data.len() as u64,
            first_inserted_id: first_inserted_id.unwrap_or(0) as u64,
        })
    }

    async fn do_update(
        &mut self,
        q: Cow<'_, UpdateStatement>,
        params: Option<&[DfValue]>,
    ) -> ReadySetResult<QueryResult<'_>> {
        trace!(table = %q.table.name, "update::access mutator");
        let mutator = self.inner.get_mut()?.get_noria_table(&q.table).await?;

        let q = q.into_owned();
        let (key, updates) = {
            trace!("update::extract schema");
            let schema = if let Some(cts) = mutator.schema() {
                cts
            } else {
                // no update on views
                unsupported!();
            };
            let coerced_params =
                utils::coerce_params(params, &SqlQuery::Update(q.clone()), schema, self.dialect)?;
            utils::extract_update(
                q,
                coerced_params.map(|p| p.into_iter()),
                schema,
                self.dialect,
            )?
        };

        trace!("update::update");
        mutator.update(key, updates).await?;
        trace!("update::complete");
        // TODO: return meaningful fields for (num_rows_updated, last_inserted_id) rather than
        // hardcoded (1,0)
        Ok(QueryResult::Update {
            num_rows_updated: 1,
            last_inserted_id: 0,
        })
    }

    async fn do_delete<'a>(
        &'a mut self,
        q: Cow<'_, DeleteStatement>,
        params: Option<&[DfValue]>,
    ) -> ReadySetResult<QueryResult<'a>> {
        trace!(table = %q.table.name, "delete::access mutator");
        let mutator = self.inner.get_mut()?.get_noria_table(&q.table).await?;

        let q = q.into_owned();
        let key = {
            trace!("delete::extract schema");
            let schema = if let Some(cts) = mutator.schema() {
                cts
            } else {
                // no delete on views
                unsupported!();
            };
            let coerced_params =
                utils::coerce_params(params, &SqlQuery::Delete(q.clone()), schema, self.dialect)?;
            utils::extract_delete(q, coerced_params.map(|p| p.into_iter()), schema)?
        };

        trace!("delete::delete");
        mutator.delete(key).await?;
        trace!("delete::complete");
        // TODO: return meaningful fields for (num_rows_deleted, last_inserted_id) rather than
        // hardcoded (1,0)
        Ok(QueryResult::Delete {
            num_rows_deleted: 1,
        })
    }

    #[instrument(level = "info", skip(self, statement))]
    pub(crate) async fn prepare_select(
        &mut self,
        mut statement: nom_sql::SelectStatement,
        statement_id: u32,
        create_if_not_exist: bool,
        override_schema_search_path: Option<Vec<SqlIdentifier>>,
    ) -> ReadySetResult<PrepareResult> {
        // extract parameter columns *for the client*
        // note that we have to do this *before* processing the query, otherwise the
        // client will be confused about the number of parameters it's supposed to
        // give.
        let client_param_columns: Vec<_> = utils::select_statement_parameter_columns(&statement)
            .into_iter()
            .cloned()
            .collect();

        let limit_columns: Vec<_> = utils::get_limit_parameters(&statement)
            .into_iter()
            .map(|column| ColumnSchema {
                column,
                column_type: DfType::UnsignedBigInt,
                base: None,
            })
            .collect();

        trace!("select::collapse where-in clauses");
        let processed_query_params =
            rewrite::process_query(&mut statement, self.server_supports_pagination())?;

        // check if we already have this query prepared
        trace!("select::access view");
        let qname = self.get_view(&statement, true, create_if_not_exist).await?;

        // extract result schema
        trace!(qname = %qname, "select::extract schema");
        let view_failed = self.failed_views.take(&qname).is_some();
        let getter = self
            .inner
            .get_mut()?
            .get_noria_view(&qname, view_failed)
            .await?;

        let getter_schema = getter
            .schema()
            .ok_or_else(|| internal_err!("no schema for view '{}'", qname))?;

        let mut params: Vec<_> = getter_schema
            .to_cols(&client_param_columns, SchemaType::ProjectedSchema)?
            .into_iter()
            .map(|cs| {
                let mut cs = cs.clone();
                cs.column.table = Some(qname.clone());
                cs
            })
            .collect();

        trace!(id = statement_id, "select::registered");
        let ps = PreparedSelectStatement {
            name: qname,
            statement: Box::new(statement),
            processed_query_params,
        };
        self.prepared_statement_cache
            .insert(statement_id, PreparedStatement::Select(ps));

        params.extend(limit_columns);
        Ok(PrepareResult::Select {
            statement_id,
            params,
            schema: getter_schema.schema(SchemaType::ReturnedSchema).to_vec(),
        })
    }

    #[instrument(level = "debug", skip(self, event))]
    pub(crate) async fn execute_select(
        &mut self,
        ctx: ExecuteSelectContext<'_>,
        ticket: Option<Timestamp>,
        event: &mut readyset_client_metrics::QueryExecutionEvent,
    ) -> ReadySetResult<QueryResult<'_>> {
        let (qname, statement, processed_query_params, params) = match ctx {
            ExecuteSelectContext::Prepared { q_id, params } => {
                let PreparedSelectStatement {
                    name,
                    statement,
                    processed_query_params,
                } = {
                    match self.prepared_statement_cache.get(&q_id) {
                        Some(PreparedStatement::Select(ps)) => ps,
                        Some(_) => internal!(),
                        None => return Err(PreparedStatementMissing { statement_id: q_id }),
                    }
                };
                (
                    Cow::Borrowed(name),
                    Cow::Borrowed(statement.as_ref()),
                    Cow::Borrowed(processed_query_params),
                    params,
                )
            }
            ExecuteSelectContext::AdHoc {
                mut statement,
                query,
                create_if_missing,
            } => {
                verify_no_placeholders(&mut statement, query)?;
                let processed_query_params =
                    rewrite::process_query(&mut statement, self.server_supports_pagination())?;
                let name = self.get_view(&statement, false, create_if_missing).await?;
                (
                    Cow::Owned(name),
                    Cow::Owned(statement),
                    Cow::Owned(processed_query_params),
                    &[][..],
                )
            }
        };

        let view_failed = self.failed_views.take(qname.as_ref()).is_some();
        let getter = self
            .inner
            .get_mut()?
            .get_noria_view(&qname, view_failed)
            .await?;

        let res = do_read(
            getter,
            processed_query_params.as_ref(),
            params,
            statement.as_ref(),
            ticket,
            self.read_behavior,
            self.read_request_handler.as_mut(),
            event,
            self.dialect,
        )
        .await;

        if let Err(e) = res.as_ref() {
            if e.is_networking_related() || e.caused_by_view_destroyed() {
                self.failed_views.insert(qname.into_owned());
            }
        }

        res
    }

    pub(crate) async fn handle_create_view<'a>(
        &'a mut self,
        q: &nom_sql::CreateViewStatement,
    ) -> ReadySetResult<QueryResult<'a>> {
        // TODO(malte): we should perhaps check our usual caches here, rather than just blindly
        // doing a migration on ReadySet every time. On the other hand, CREATE VIEW is rare...

        info!(view = %Sensitive(&q.definition), name = %q.name, "view::create");

        let changelist = ChangeList::from_change(Change::CreateView(q.clone()), self.dialect)
            .with_schema_search_path(self.schema_search_path.clone());

        noria_await!(
            self.inner.get_mut()?,
            self.inner.get_mut()?.noria.extend_recipe(changelist)
        )?;
        Ok(QueryResult::Empty)
    }
}

/// Verifies that there are no placeholder parameters in the given SELECT statement (i.e. ? or $N),
/// returning `Ok(())` if none are found, or an `InvalidQuery` error if there are any placeholders
/// present in the statement.
fn verify_no_placeholders(statement: &mut SelectStatement, query: &str) -> ReadySetResult<()> {
    struct PlaceholderFoundVisitor;

    impl<'ast> VisitorMut<'ast> for PlaceholderFoundVisitor {
        type Error = ();

        fn visit_literal(&mut self, literal: &'ast mut Literal) -> Result<(), Self::Error> {
            if matches!(literal, Literal::Placeholder(_)) {
                Err(())
            } else {
                Ok(())
            }
        }
    }

    if PlaceholderFoundVisitor
        .visit_select_statement(statement)
        .is_err()
    {
        Err(ReadySetError::InvalidQuery(query.to_string()))
    } else {
        Ok(())
    }
}

/// Build a [`ViewQuery`] for performing a lookup of the given `q` with the given `raw_keys`,
/// provided `getter_schema` and `key_map` from the [`View`] itself.
#[allow(clippy::too_many_arguments)]
fn build_view_query(
    getter_schema: &ViewSchema,
    key_map: &[(ViewPlaceholder, KeyColumnIdx)],
    processed_query_params: &ProcessedQueryParams,
    params: &[DfValue],
    q: &nom_sql::SelectStatement,
    ticket: Option<Timestamp>,
    read_behavior: ReadBehavior,
    dialect: Dialect,
) -> ReadySetResult<ViewQuery> {
    let projected_schema = getter_schema.schema(SchemaType::ProjectedSchema);

    let (limit, offset) = processed_query_params.limit_offset_params(params)?;

    let mut raw_keys = processed_query_params.make_keys(params)?;

    let mut key_types = getter_schema.col_types(
        key_map.iter().map(|(_, key_column_idx)| *key_column_idx),
        SchemaType::ProjectedSchema,
    )?;

    trace!("select::lookup");
    let bogo = vec![vec1![DfValue::from(0i32)].into()];
    let mut binops = utils::get_select_statement_binops(q);
    let mut filter_op_idx = None;
    let mut filters = binops
        .iter()
        .enumerate()
        .filter(|(_, (_, binop))| matches!(binop, BinaryOperator::Like | BinaryOperator::ILike))
        .map(|(idx, (col, op))| -> ReadySetResult<_> {
            let key = raw_keys.drain(0..1).next().ok_or(ReadySetError::EmptyKey)?;
            if !raw_keys.is_empty() {
                unsupported!(
                    "LIKE/ILIKE not currently supported for more than one lookup key at a time"
                );
            }
            let column = projected_schema
                .iter()
                .position(|x| x.column.name == col.name)
                .ok_or_else(|| ReadySetError::NoSuchColumn(col.name.to_string()))?;

            let key_type = key_types.remove(idx);
            let value = key[idx].coerce_to(key_type, &DfType::Unknown)?; // No from_ty, key values are literals

            if !key.is_empty() {
                // the LIKE/ILIKE isn't our only key, add the rest back to `keys`
                raw_keys.push(key);
            }

            filter_op_idx = Some(idx);

            // LIKE/ILIKE resolve to bool
            Ok(DfExpr::Op {
                left: Box::new(DfExpr::Column {
                    index: column,
                    ty: key_type.clone(),
                }),
                op: DfBinaryOperator::from_sql_op(*op, dialect),
                right: Box::new(DfExpr::Literal {
                    val: value,
                    ty: key_type.clone(),
                }),
                ty: DfType::Bool,
            })
        })
        .collect::<Result<Vec<_>, _>>()?;

    if let Some(filter_op_idx) = filter_op_idx {
        // if we're using a column for a post-lookup filter, remove it from our list of binops
        // so we can use the remaining list for our keys
        binops.remove(filter_op_idx);
    }

    let keys = if raw_keys.is_empty() {
        bogo
    } else {
        let mut unique_binops = binops.iter().map(|(_, b)| *b).unique();
        let binop_to_use = unique_binops.next().unwrap_or(BinaryOperator::Equal);
        let mixed_binops = unique_binops.next().is_some();

        let key_types: HashMap<usize, &DfType> = key_map
            .iter()
            .zip(key_types)
            .map(|((_, key_column_idx), key_type)| (*key_column_idx, key_type))
            .collect();

        raw_keys
            .into_iter()
            .map(|key| {
                let mut k = vec![];
                let mut bounds: Option<(Vec<DfValue>, Vec<DfValue>)> = if mixed_binops {
                    Some((vec![], vec![]))
                } else {
                    None
                };
                for (view_placeholder, key_column_idx) in key_map {
                    match view_placeholder {
                        ViewPlaceholder::Generated => continue,
                        ViewPlaceholder::OneToOne(idx) => {
                            // If the key isn't in the list of key types, that means it was removed
                            // during post-read filter construction (for LIKE/ILIKE) above, so just
                            // skip it in the lookup key.
                            let key_type = match key_types.get(key_column_idx) {
                                Some(&ty) => ty,
                                None => continue,
                            };

                            // parameter numbering is 1-based, but vecs are 0-based, so subtract 1
                            // also, no from_ty since the key value is a literal
                            let value = key[*idx - 1].coerce_to(key_type, &DfType::Unknown)?;

                            let make_op = |op: DfBinaryOperator| DfExpr::Op {
                                left: Box::new(DfExpr::Column {
                                    index: *key_column_idx,
                                    ty: key_type.clone(),
                                }),
                                op,
                                right: Box::new(DfExpr::Literal {
                                    val: value.clone(),
                                    ty: key_type.clone(),
                                }),
                                ty: DfType::Bool, // TODO: infer type
                            };

                            if let Some((lower_bound, upper_bound)) = &mut bounds {
                                let binop = binops[*idx - 1].1;
                                match binop {
                                    BinaryOperator::Like
                                    | BinaryOperator::NotLike
                                    | BinaryOperator::ILike
                                    | BinaryOperator::NotILike => {
                                        internal!("Already should have matched on LIKE above")
                                    }

                                    BinaryOperator::Equal => {
                                        lower_bound.push(value.clone());
                                        upper_bound.push(value);
                                    }
                                    BinaryOperator::GreaterOrEqual => {
                                        filters.push(make_op(DfBinaryOperator::GreaterOrEqual));
                                        lower_bound.push(value);
                                        upper_bound.push(DfValue::Max);
                                    }
                                    BinaryOperator::LessOrEqual => {
                                        filters.push(make_op(DfBinaryOperator::LessOrEqual));
                                        lower_bound.push(DfValue::None); // NULL is the minimum DfValue
                                        upper_bound.push(value);
                                    }
                                    BinaryOperator::Greater => {
                                        filters.push(make_op(DfBinaryOperator::Greater));
                                        lower_bound.push(value);
                                        upper_bound.push(DfValue::Max);
                                    }
                                    BinaryOperator::Less => {
                                        filters.push(make_op(DfBinaryOperator::Less));
                                        lower_bound.push(DfValue::None); // NULL is the minimum DfValue
                                        upper_bound.push(value);
                                    }
                                    op => unsupported!(
                                        "Unsupported binary operator in query: `{}`",
                                        op
                                    ),
                                }
                            } else {
                                if !k.is_empty() && binop_to_use != BinaryOperator::Equal {
                                    filters.push(make_op(DfBinaryOperator::from_sql_op(
                                        binop_to_use,
                                        dialect,
                                    )));
                                }
                                k.push(value);
                            }
                        }
                        ViewPlaceholder::Between(lower_idx, upper_idx) => {
                            let key_type = key_types[key_column_idx];

                            // parameter numbering is 1-based, but vecs are 0-based, so subtract 1
                            // also, no from_ty since the key value is a literal
                            let lower_value =
                                key[*lower_idx - 1].coerce_to(key_type, &DfType::Unknown)?;
                            let upper_value =
                                key[*upper_idx - 1].coerce_to(key_type, &DfType::Unknown)?;
                            let (lower_key, upper_key) =
                                bounds.get_or_insert_with(Default::default);
                            lower_key.push(lower_value);
                            upper_key.push(upper_value);
                        }
                        ViewPlaceholder::PageNumber {
                            offset_placeholder,
                            limit,
                        } => {
                            // parameter numbering is 1-based, but vecs are 0-based, so subtract 1
                            // offset parameters should always be a BigInt
                            // also, no from_ty since the key value is a literal
                            let offset: u64 = key[*offset_placeholder - 1]
                                .coerce_to(&DfType::BigInt, &DfType::Unknown)?
                                .try_into()?;
                            if offset % *limit != 0 {
                                unsupported!(
                                    "OFFSET must currently be an integer multiple of LIMIT"
                                );
                            }
                            let page_number = offset / *limit;
                            k.push(page_number.into());
                        }
                    };
                }

                if let Some((lower, upper)) = bounds {
                    debug_assert!(k.is_empty());
                    Ok(KeyComparison::Range((
                        Bound::Included(lower.try_into()?),
                        Bound::Included(upper.try_into()?),
                    )))
                } else {
                    KeyComparison::from_key_and_operator(k, binop_to_use)
                }
            })
            .collect::<ReadySetResult<Vec<_>>>()?
    };

    trace!(?keys, ?filters, "Built view query");

    Ok(ViewQuery {
        key_comparisons: keys,
        block: read_behavior.is_blocking(),
        filter: filters.into_iter().reduce(|expr1, expr2| DfExpr::Op {
            left: Box::new(expr1),
            op: DfBinaryOperator::And,
            right: Box::new(expr2),
            ty: DfType::Bool, // AND is a boolean operator
        }),
        limit,
        offset,
        timestamp: ticket,
    })
}

/// Run the supplied [`SelectStatement`] on the supplied [`View`]
/// Assumption: the [`View`] was created for that specific [`SelectStatement`]
#[allow(clippy::needless_lifetimes)] // clippy erroneously thinks the timelife can be elided
#[allow(clippy::too_many_arguments)]
async fn do_read<'a>(
    getter: &'a mut View,
    processed_query_params: &ProcessedQueryParams,
    params: &[DfValue],
    q: &nom_sql::SelectStatement,
    ticket: Option<Timestamp>,
    read_behavior: ReadBehavior,
    read_request_handler: Option<&'a mut ReadRequestHandler>,
    event: &mut readyset_client_metrics::QueryExecutionEvent,
    dialect: Dialect,
) -> ReadySetResult<QueryResult<'a>> {
    let (limit, _) = processed_query_params.limit_offset_params(params)?;
    if limit == Some(0) {
        return short_circuit_empty_resultset(getter).await;
    }

    let vq = build_view_query(
        getter
            .schema()
            .ok_or_else(|| internal_err!("No schema for view"))?,
        getter.key_map(),
        processed_query_params,
        params,
        q,
        ticket,
        read_behavior,
        dialect,
    )?;

    event.num_keys = Some(vq.key_comparisons.len() as _);

    let data = if let Some(rh) = read_request_handler {
        let request = readyset::Tagged::from(ReadQuery::Normal {
            target: ReaderAddress {
                node: *getter.node(),
                name: getter.name().clone(),
                shard: 0,
            },
            query: vq.clone(),
        });

        // Query the local reader if it is a read query, otherwise default to the traditional
        // View API.
        let tag = request.tag;
        if let ReadQuery::Normal { target, query } = request.v {
            // Issue a normal read query returning the raw unserialized results.
            let result = match rh.handle_normal_read_query(tag, target, query, true) {
                CallResult::Immediate(result) => result?,
                CallResult::Async(chan) => chan.await?,
            };

            result
                .v
                .into_normal()
                .ok_or_else(|| internal_err!("Unexpected response type from reader service"))??
                .into_results()
                .ok_or(ReadySetError::ReaderMissingKey)?
                .pop()
                .ok_or_else(|| internal_err!("Expected a single result set for local reader"))?
                .into_unserialized()
                .expect("Requested raw result")
        } else {
            getter.raw_lookup(vq).await?
        }
    } else {
        getter.raw_lookup(vq).await?
    };

    event.cache_misses = data.total_stats().map(|s| s.cache_misses);

    trace!("select::complete");

    Ok(QueryResult::from_iter(
        SelectSchema {
            // TODO(vlad): looks like poor `use_bogo` is unused except in js? Should just remove it.
            use_bogo: false,
            schema: Cow::Borrowed(getter.schema().unwrap().schema(SchemaType::ReturnedSchema)), /* Safe because we already unwrapped above */
            columns: Cow::Borrowed(getter.columns()),
        },
        data,
    ))
}

#[cfg(test)]
mod tests {
    use nom_sql::Dialect;

    use super::*;

    mod view_cache {
        use nom_sql::{parse_select_statement, Dialect, Relation};

        use super::*;

        #[test]
        fn register_and_remove_statement() {
            let global = Arc::new(RwLock::new(HashMap::new()));
            let mut view_cache = ViewCache::new(global);

            let name = Relation::from("test_statement_name");
            let statement = parse_select_statement(Dialect::MySQL, "SELECT a_col FROM t1").unwrap();
            let view_request = ViewCreateRequest::new(statement, vec!["s1".into()]);

            view_cache.register_statement(&name, view_request.clone());
            let retrieved_request = view_cache.view_create_request_from_name(&name);
            assert_eq!(Some(view_request), retrieved_request);

            view_cache.remove_statement(&name);
            let retrieved_request = view_cache.view_create_request_from_name(&name);
            assert_eq!(None, retrieved_request);
        }

        #[test]
        fn clear() {
            let global = Arc::new(RwLock::new(HashMap::new()));
            let mut view_cache = ViewCache::new(global.clone());

            let statement1 = parse_select_statement(Dialect::MySQL, "SELECT a FROM t1").unwrap();
            let statement2 = parse_select_statement(Dialect::MySQL, "SELECT b FROM t2").unwrap();
            let create_request_1 = ViewCreateRequest::new(statement1, vec!["schema1".into()]);
            let create_request_2 = ViewCreateRequest::new(statement2, vec!["schema2".into()]);

            view_cache.register_statement(&"q1".into(), create_request_1.clone());
            view_cache.register_statement(&"q2".into(), create_request_2.clone());

            assert_eq!(
                view_cache.view_create_request_from_name(&"q1".into()),
                Some(create_request_1)
            );
            assert_eq!(
                view_cache.view_create_request_from_name(&"q2".into()),
                Some(create_request_2)
            );

            view_cache.clear();

            assert_eq!(view_cache.view_create_request_from_name(&"q1".into()), None);
            assert_eq!(view_cache.view_create_request_from_name(&"q2".into()), None);
            assert!(global.read().unwrap().is_empty());
        }
    }

    mod build_view_query {
        use dataflow_expression::Dialect as DfDialect;
        use lazy_static::lazy_static;
        use nom_sql::{parse_query, Column, Dialect, SelectStatement};
        use readyset::ColumnBase;

        use super::*;

        lazy_static! {
            static ref SCHEMA: ViewSchema = ViewSchema::new(
                vec![
                    ColumnSchema {
                        column: Column {
                            name: "x".into(),
                            table: Some("t".into()),
                        },
                        column_type: DfType::Int,
                        base: Some(ColumnBase {
                            table: "t".into(),
                            column: "x".into(),
                            constraints: vec![],
                        }),
                    },
                    ColumnSchema {
                        column: Column {
                            name: "y".into(),
                            table: Some("t".into()),
                        },
                        column_type: DfType::DEFAULT_TEXT,
                        base: Some(ColumnBase {
                            table: "t".into(),
                            column: "y".into(),
                            constraints: vec![],
                        }),
                    }
                ],
                vec![
                    ColumnSchema {
                        column: Column {
                            name: "x".into(),
                            table: Some("t".into()),
                        },
                        column_type: DfType::Int,
                        base: Some(ColumnBase {
                            table: "t".into(),
                            column: "x".into(),
                            constraints: vec![],
                        }),
                    },
                    ColumnSchema {
                        column: Column {
                            name: "y".into(),
                            table: Some("t".into()),
                        },
                        column_type: DfType::DEFAULT_TEXT,
                        base: Some(ColumnBase {
                            table: "t".into(),
                            column: "y".into(),
                            constraints: vec![],
                        }),
                    }
                ],
            );
        }

        fn parse_select_statement(src: &str) -> SelectStatement {
            match parse_query(Dialect::MySQL, src).unwrap() {
                SqlQuery::Select(stmt) => stmt,
                _ => panic!("Unexpected query type"),
            }
        }

        fn make_build_query(
            query: &str,
            key_map: &[(ViewPlaceholder, KeyColumnIdx)],
            params: &[DfValue],
            dialect: Dialect,
        ) -> ViewQuery {
            let dataflow_dialect = match dialect {
                Dialect::MySQL => DfDialect::DEFAULT_MYSQL,
                Dialect::PostgreSQL => DfDialect::DEFAULT_POSTGRESQL,
            };

            let mut q = parse_select_statement(query);
            let pp = rewrite::process_query(&mut q, true).unwrap();
            build_view_query(
                &*SCHEMA,
                key_map,
                &pp,
                params,
                &q,
                None,
                ReadBehavior::Blocking,
                dataflow_dialect,
            )
            .unwrap()
        }

        #[test]
        fn simple_point_lookup() {
            let query = make_build_query(
                "SELECT t.x FROM t WHERE t.x = $1",
                &[(ViewPlaceholder::OneToOne(1), 0)],
                &[DfValue::from(1)],
                Dialect::MySQL,
            );

            assert!(query.filter.is_none());
            assert_eq!(
                query.key_comparisons,
                vec![KeyComparison::from(vec1![DfValue::from(1)])]
            );
        }

        #[test]
        fn single_between() {
            let query = make_build_query(
                "SELECT t.x FROM t WHERE t.x BETWEEN $1 AND $2",
                &[(ViewPlaceholder::Between(1, 2), 0)],
                &[DfValue::from(1), DfValue::from(2)],
                Dialect::MySQL,
            );

            assert!(query.filter.is_none());
            assert_eq!(
                query.key_comparisons,
                vec![KeyComparison::from_range(
                    &(vec1![DfValue::from(1)]..=vec1![DfValue::from(2)])
                )]
            );
        }

        #[test]
        fn ilike_and_equality() {
            let query = make_build_query(
                "SELECT t.x FROM t WHERE t.x = $1 AND t.y ILIKE $2",
                &[
                    (ViewPlaceholder::OneToOne(1), 0),
                    (ViewPlaceholder::OneToOne(2), 1),
                ],
                &[DfValue::from(1), DfValue::from("%a%")],
                Dialect::MySQL,
            );

            assert_eq!(
                query.filter,
                Some(DfExpr::Op {
                    left: Box::new(DfExpr::Column {
                        index: 1,
                        ty: DfType::DEFAULT_TEXT
                    }),
                    op: DfBinaryOperator::ILike,
                    right: Box::new(DfExpr::Literal {
                        val: DfValue::from("%a%"),
                        ty: DfType::DEFAULT_TEXT
                    }),
                    ty: DfType::Bool,
                })
            );
        }

        #[test]
        fn mixed_equal_and_inclusive() {
            let query = make_build_query(
                "SELECT t.x FROM t WHERE t.x >= $1 AND t.y = $2",
                &[
                    (ViewPlaceholder::OneToOne(2), 1),
                    (ViewPlaceholder::OneToOne(1), 0),
                ],
                &[DfValue::from(1), DfValue::from("a")],
                Dialect::MySQL,
            );

            assert_eq!(
                query.key_comparisons,
                vec![KeyComparison::from_range(
                    &(vec1![DfValue::from("a"), DfValue::from(1)]
                        ..=vec1![DfValue::from("a"), DfValue::Max])
                )]
            );
        }

        #[test]
        fn mixed_equal_and_between() {
            let query = make_build_query(
                "SELECT t.x FROM t WHERE t.x BETWEEN $1 AND $2 AND t.y = $3",
                &[
                    (ViewPlaceholder::OneToOne(3), 1),
                    (ViewPlaceholder::Between(1, 2), 0),
                ],
                &[DfValue::from(1), DfValue::from(2), DfValue::from("a")],
                Dialect::MySQL,
            );

            assert!(query.filter.is_none());
            assert_eq!(
                query.key_comparisons,
                vec![KeyComparison::from_range(
                    &(vec1![DfValue::from("a"), DfValue::from(1)]
                        ..=vec1![DfValue::from("a"), DfValue::from(2)])
                )]
            );
        }

        #[test]
        fn mixed_equal_and_exclusive() {
            let query = make_build_query(
                "SELECT t.x FROM t WHERE t.x > $1 AND t.y = $2",
                &[
                    (ViewPlaceholder::OneToOne(2), 1),
                    (ViewPlaceholder::OneToOne(1), 0),
                ],
                &[DfValue::from(1), DfValue::from("a")],
                Dialect::MySQL,
            );

            assert_eq!(
                query.filter,
                Some(DfExpr::Op {
                    left: Box::new(DfExpr::Column {
                        index: 0,
                        ty: DfType::Int
                    }),
                    op: DfBinaryOperator::Greater,
                    right: Box::new(DfExpr::Literal {
                        val: 1.into(),
                        ty: DfType::Int
                    }),
                    ty: DfType::Bool,
                })
            );
            assert_eq!(
                query.key_comparisons,
                vec![KeyComparison::from_range(
                    &(vec1![DfValue::from("a"), DfValue::from(1)]
                        ..=vec1![DfValue::from("a"), DfValue::Max])
                )]
            );
        }

        #[test]
        fn compound_range() {
            let query = make_build_query(
                "SELECT t.x FROM t WHERE t.x > $1 AND t.y > $2",
                &[
                    (ViewPlaceholder::OneToOne(1), 0),
                    (ViewPlaceholder::OneToOne(2), 1),
                ],
                &[DfValue::from(1), DfValue::from("a")],
                Dialect::MySQL,
            );

            assert_eq!(
                query.filter,
                Some(DfExpr::Op {
                    left: Box::new(DfExpr::Column {
                        index: 1,
                        ty: DfType::DEFAULT_TEXT
                    }),
                    op: DfBinaryOperator::Greater,
                    right: Box::new(DfExpr::Literal {
                        val: "a".into(),
                        ty: DfType::DEFAULT_TEXT
                    }),
                    ty: DfType::Bool
                })
            );

            assert_eq!(
                query.key_comparisons,
                vec![KeyComparison::Range((
                    Bound::Excluded(vec1![DfValue::from(1), DfValue::from("a")]),
                    Bound::Unbounded
                ))]
            );
        }

        #[test]
        fn paginated_with_key() {
            let query = make_build_query(
                "SELECT t.x FROM t WHERE t.x = $1 ORDER BY t.y ASC LIMIT 3 OFFSET $2",
                &[
                    (ViewPlaceholder::OneToOne(1), 0),
                    (
                        ViewPlaceholder::PageNumber {
                            offset_placeholder: 2,
                            limit: 3,
                        },
                        1,
                    ),
                ],
                &[DfValue::from(1), DfValue::from(3)],
                Dialect::MySQL,
            );

            assert_eq!(query.filter, None);

            assert_eq!(
                query.key_comparisons,
                vec![vec1![
                    DfValue::from(1),
                    DfValue::from(1) // page 2
                ]
                .into()]
            );
        }
    }

    #[test]
    fn placeholder_verification_good() {
        let query = "SELECT n FROM t WHERE c = 123;";
        let parsed_query = nom_sql::parse_query(Dialect::MySQL, query).unwrap();

        match parsed_query {
            SqlQuery::Select(mut select) => {
                assert!(verify_no_placeholders(&mut select, query).is_ok())
            }
            _ => panic!(),
        }
    }

    #[test]
    fn placeholder_verification_bad() {
        let query = "SELECT n FROM t WHERE c = ?;";
        let parsed_query = nom_sql::parse_query(Dialect::MySQL, query).unwrap();

        match parsed_query {
            SqlQuery::Select(mut select) => {
                assert!(verify_no_placeholders(&mut select, query).is_err())
            }
            _ => panic!(),
        }
    }
}
